#include "aimtrans.h"

struct buddy_clean_data
{
    ati ti;
    xmlnode node;
};

result at_buddy_pending_clean(void *arg)
{
    struct buddy_clean_data *bcd;
    ati ti;
    char *user;
    xmlnode node;

    bcd = (struct buddy_clean_data *)arg;
    node = bcd->node;
    ti = bcd->ti;
    user = xmlnode_get_attrib(node, "user");

    log_debug(ZONE, "[AT] Cleaning pending for %s: %s", user, xmlnode2str(node));

    xhash_zap(ti->pending__buddies, user);

    xmlnode_free(node);

    return r_UNREG;
}

char *at_buddy_buildlist(at_session s, jid from)
{
    at_buddy new;
    char *list;
	char *blist;
	spool sp;
	int found=0;
    pool p, bp;
    xmlnode x;
    xmlnode item;
    xmlnode msg;

    x = at_xdb_get(s->ti, from, AT_NS_ROSTER);

    if(x == NULL)
    {
        xmlnode msg;
        xmlnode body;
        
        msg = xmlnode_new_tag("message");
        xmlnode_put_attrib(msg, "to", jid_full(from));
        xmlnode_put_attrib(msg, "from", s->ti->i->id);
        body = xmlnode_insert_tag(msg, "body");
        xmlnode_insert_cdata(body, "The AIM Transport now stores your contacts on the server to better provide presence information.  It is highly suggested that you send presence to all of your AIM contacts and then logoff of the server.  This will only need to be done once to save the information on the server.", -1);
        deliver(dpacket_new(msg), s->ti->i);
                
        x = xhash_get(s->ti->pending__buddies, jid_full(from));
        if(x == NULL)
            return NULL;
    }
    
    p = pool_new();
	sp = spool_new(p);

    for(item = xmlnode_get_firstchild(x); item != NULL; 
                    item = xmlnode_get_nextsibling(item))
    {
        char *sn;

        sn = xmlnode_get_attrib(item, "name");

 		log_debug(ZONE, "[AIM] Adding a buddy %s from xdb", sn);

        spooler(sp, sn, "&", sp);
        new = pmalloco(s->p, sizeof(_at_buddy));
        new->full = jid_new(s->p, s->ti->i->id);
        jid_set(new->full, sn, JID_USER);
        new->last = xmlnode_new_tag_pool(s->p,"query");
        xmlnode_put_attrib(new->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(new->last,"stamp",jutil_timestamp()); 

        found = 1;
        xhash_put(s->buddies, new->full->user, new);
    }

    list = spool_print(sp);
    if(list)
    {
        blist = strdup(list);
    } else {
        blist = NULL;
    }
    log_debug(ZONE, "[AT] Buddylist generation complete");

    pool_free(p);

    return blist;
}

int at_buddy_subscribe(ati ti, jpacket jp)
{
    xmlnode dup,dup2;
    at_session s=at_session_find_by_jid(ti, jp->from);

    /* check for a valid session */
    if(s==NULL)
    {
        xmlnode err,error;
        /* if no session, and subscribing to a buddy, not aim-t */
        err=xmlnode_new_tag("message");
        xmlnode_put_attrib(err,"type","error");
        xmlnode_put_attrib(err,"from",ti->i->id);
        xmlnode_put_attrib(err,"to",jid_full(jp->from));
        error=xmlnode_insert_tag(err,"error");
        xmlnode_insert_cdata(error,"Cannot Subscribe to a AIM Buddy without a registration",-1);
        xmlnode_put_attrib(error,"code","407");
        deliver(dpacket_new(err), ti->i);

        return 0;
    }
    /* XXX
	 * Create the duplicate for the subscribe notice
	 * Construct a subscribed notification
	 * 
	 * I had to do these in this odd order due to memory concerns
	 * Dunno why I didn't just make two dups.
	 *
	 */
	dup = xmlnode_dup(jp->x);
	dup2 = xmlnode_dup(jp->x);

	xmlnode_put_attrib(dup, "to", jid_full(jp->from));
	xmlnode_put_attrib(dup, "from", jid_full(jp->to));
	xmlnode_put_attrib(dup, "type", "subscribed");
	log_debug(ZONE, "[AIM] Sending subscribed notice\n");
    deliver(dpacket_new(dup), ti->i);

	/* Construct the subscribe back */
	xmlnode_put_attrib(dup2, "type", "subscribe");
    xmlnode_put_attrib(dup2, "to", jid_full(jp->from));
    xmlnode_put_attrib(dup2, "from", jid_full(jp->to));
	log_debug(ZONE, "[AIM] Asking for a subscribe\n");
    deliver(dpacket_new(dup2), ti->i);

	return 0;
}

int at_buddy_add(ati ti, jpacket jp)
{
	at_session s;
	at_buddy buddy;
    pool p;
	int newbud=0;
    struct buddy_clean_data *bcd;
	
    log_debug(ZONE, "[AIM] Going to add a buddy\n");
	s = at_session_find_by_jid(ti, jp->from);
	if(s==NULL || s->loggedin == 0) {
        xmlnode cur, item;
        
        cur = xhash_get(ti->pending__buddies, jid_full(jp->from));
        if(cur == NULL)
        {
            cur = xmlnode_new_tag("buddies");
            xmlnode_put_attrib(cur, "user", jid_full(jp->from));
            bcd = pmalloco(xmlnode_pool(cur), sizeof(struct buddy_clean_data));
            bcd->node = cur;
            bcd->ti = ti;
            register_beat(30, at_buddy_pending_clean, bcd);
        }
        item = xmlnode_insert_tag(cur, "item");
        xmlnode_put_attrib(item, "name", jp->to->user);
        log_debug(ZONE, "[AT] Pending for %s: %s", jid_full(jp->from), xmlnode2str(cur));
        xhash_put(ti->pending__buddies, jid_full(jp->from), cur);
        return 0;
	}
    
    if(xhash_get(s->buddies, jp->to->user) == NULL)
    {
		log_debug(ZONE, "[AIM] Adding buddy %s to %s\n", jp->to->user, jid_full(jp->from));
        buddy = pmalloco(s->p, sizeof(_at_buddy));
        buddy->full=jid_new(s->p, jid_full(jp->to));
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());
        xhash_put(s->buddies, buddy->full->user, buddy);
        at_buddy_subscribe(ti, jp);
    }
    aim_add_buddy(s->ass, aim_getconn_type(s->ass, AIM_CONN_TYPE_BOS),
                  jp->to->user);
    
    return 0;
}

int at_parse_oncoming(struct aim_session_t *ass, 
                      struct command_rx_struct *command, ...)
{

    xmlnode away;
    xmlnode status;
	xmlnode x;
	at_session s;
    at_buddy buddy;
    ati ti;
    jpacket jp;
    char *msg;
    struct aim_userinfo_s *userinfo;

    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

    log_debug(ZONE, "Oncoming buddy %s", userinfo->sn);

	s = (at_session)ass->aux_data;
    ti = s->ti;

    buddy = xhash_get(s->buddies, userinfo->sn);
    if(buddy == NULL)
    {
        jid jtmp;

        buddy = pmalloco(s->p, (sizeof(_at_buddy)));
        buddy->full = jid_new(s->p, ti->i->id);
        jid_set(buddy->full, userinfo->sn, JID_USER);
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());

        xhash_put(s->buddies, buddy->full->user, buddy);
    }

	x = xmlnode_new_tag("presence");
	xmlnode_put_attrib(x, "to", jid_full(s->cur));
	xmlnode_put_attrib(x, "from", ti->i->id);
	jp = jpacket_new(x);
	jid_set(jp->from, at_normalize(userinfo->sn), JID_USER);
	xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));
    if(userinfo->flags&AIM_FLAG_AWAY)
    {
        away = xmlnode_insert_tag(x, "show");
        xmlnode_insert_cdata(away, "away", 4);
    }
    status = xmlnode_insert_tag(x, "status");
    msg = pmalloco(xmlnode_pool(x), 24);
    memset(msg, '\0', 24);
    sprintf(msg, "Idle %d Minutes", (userinfo->idletime));
    xmlnode_insert_cdata(status, msg, -1);
    deliver(dpacket_new(jp->x), ti->i);

	return 1;
}

int at_parse_offgoing(struct aim_session_t *ass, 
                      struct command_rx_struct *command, ...)
{

	xmlnode x;
	jpacket jp;
    at_buddy buddy;
	at_session s;
    ati ti;
    char *sn;
    
	s = (at_session)ass->aux_data;
    ti = s->ti;
    sn = &(command->data[11]);
    buddy = (at_buddy)xhash_get(s->buddies, sn);
    if(buddy == NULL)
    {
        jid jtmp;

        buddy = pmalloco(s->p, (sizeof(_at_buddy)));
        buddy->full = jid_new(s->p, ti->i->id);
        jid_set(buddy->full, sn, JID_USER);
        buddy->last = xmlnode_new_tag_pool(s->p,"query");
        xmlnode_put_attrib(buddy->last,"xmlns",NS_LAST);
        xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());

        xhash_put(s->buddies, buddy->full->user, buddy);
    }

    xmlnode_put_attrib(buddy->last,"stamp",jutil_timestamp());    

	x = xmlnode_new_tag("presence");
	xmlnode_put_attrib(x, "to", jid_full(s->cur));
	xmlnode_put_attrib(x, "from", ti->i->id);
	xmlnode_put_attrib(x, "type","unavailable");
	jp = jpacket_new(x);
	jid_set(jp->from, at_normalize(sn), JID_USER);
	xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));

    deliver(dpacket_new(jp->x), ti->i);

	return 1;
}

int at_parse_evilnotify(struct aim_session_t *sess, 
                              struct command_rx_struct *command, ...)
{
    va_list ap;
    char *sn;
    xmlnode x;
    xmlnode body;
    jpacket jp;
    at_session s;
    ati ti;
    char msg[100];

    va_start(ap, command);
    sn = va_arg(ap, char *);
    va_end(ap);

    memset(&msg, '\0', 100);
    snprintf((char *)&msg, 100, "Warning from: %s", sn);

    s = (at_session)sess->aux_data;
    ti = s->ti;

	x = xmlnode_new_tag("message");
	xmlnode_put_attrib(x, "to", jid_full(s->cur));
	xmlnode_put_attrib(x, "from", jid_full(s->from));
	xmlnode_put_attrib(x, "type","error");
	body = xmlnode_insert_tag(x, "error");
    xmlnode_insert_cdata(body, (char *)&msg, strlen((char *)&msg));

    jp = jpacket_new(x);
    deliver(dpacket_new(jp->x), ti->i);

    return 1;
}

int at_parse_userinfo(struct aim_session_t *sess, 
                            struct command_rx_struct *command, ...)
{
    /* XXX Do something here? */
    return 1;
}


int at_parse_locerr(struct aim_session_t *sess, 
                          struct command_rx_struct *command, ...)
{
/* XXX Do we want this?
    va_list ap;
    char *destsn;
    unsigned short reason;
    char *sn;
    xmlnode x;
    xmlnode body;
    jpacket jp;
    session s;
    char msg[1024];

    va_start(ap, command);
    destsn = va_arg(ap, char *);
    reason = va_arg(ap, unsigned short);
    va_end(ap);

    memset(&msg, '\0', 1024);
    snprintf("User information for %s unavailable (reason 0x%04x: %s)\n", destsn, reason, (reason<msgerrreasonslen)?msgerrreasons[reason]:"unknown");

    s = (session)ass->aux_data;

	x = xmlnode_new_tag("message");
	xmlnode_put_attrib(x, "to", jid_full(s->j));
	xmlnode_put_attrib(x, "from", aimtrans__instance->id);
	xmlnode_put_attrib(x, "type","error");
	body = xmlnode_insert_tag(x, "error");
    xmlnode_insert_cdata(body, &msg, strlen(&msg));

    jp = jpacket_new(x);
	ehandler_send(ehandler_at, jp->x, s->j->server);
*/  
    return 1;
}
