#include "aimtrans.h"

int at_server_pres(ati ti, jpacket jp)
{
    xmlnode x;

    switch(jpacket_subtype(jp))
    {
        case JPACKET__AVAILABLE:
            if(jp->to->resource == NULL)
            {
                return at_buddy_add(ti, jp);
            } else {
                log_debug(ZONE, "[AT] Starting a new session!\n");
                return at_auth_user(ti, jp);
            }
        case JPACKET__UNAVAILABLE:
            /* XXX These two depend on sessions so we'll do those first  */
            /* XXX If we get avail pres to a user and we have no session 
             *     then we need to put it into a list for when the session
             *     becomes active.  Client side buddy lists are such a flaw
             *     in good IM design.
             */
            log_debug(ZONE, "[AT] Unavailable sent to server");
        case JPACKET__PROBE:
            /* XXX Still not sure if we want to handle this.
             *     What are the pros and cons of doing this?
             */
        default:
            jutil_error(jp->x, TERROR_NOTACCEPTABLE);
            deliver(dpacket_new(jp->x), ti->i);
            return 1;
    }

    return 0;
}

int at_session_pres(at_session s, jpacket jp)
{
    ati ti;
    xmlnode x;
    jid j;

    ti = s->ti;

    switch(jpacket_subtype(jp))
    {
        case JPACKET__PROBE:
            log_debug(ZONE, "[AT] Probed, no logical way to handle");
            return 0;
        case JPACKET__UNAVAILABLE:
            log_debug(ZONE, "[AT] Unavailabe sent to session");
            /* insert into ppdb, check to see if we should go offline */
            s->at_PPDB=ppdb_insert(s->at_PPDB,jp->from,jp->x);
            j=jid_new(jp->p,jid_full(jp->from));
            jid_set(j,NULL,JID_RESOURCE);
            log_debug(ZONE, "[AT] Checking at_PPDB for %s", jid_full(j));
            x=ppdb_primary(s->at_PPDB,j);
            if(x!=NULL)
            {
                /* not last resource, don't go offline yet */
                log_debug(ZONE,"[AT] active resources, not ending session");
                s->cur = jid_new(s->p,xmlnode_get_attrib(x, "from"));
                return 0;
            }
			log_debug(ZONE, "[AT] Telling the session to end!");
            s->exit_flag = 1;
            xmlnode_free(jp->x);
            return 1;
		case JPACKET__AVAILABLE:
                s->at_PPDB=ppdb_insert(s->at_PPDB,jp->from,jp->x);
				log_debug(ZONE, "[AT] Starting a new session!");
                s=at_session_find_by_jid(ti, jp->from);
                log_debug(ZONE, "[AT] S is: %X", s);
				if((s) && (xmlnode_get_tag(jp->x, "show") != NULL))
                {
                    /* why are we returning here? 
                    * this doesn't seem like the proper way
                    * to handle this...
                    */   
                    char *awaymsg;
                    char profile[]="";

                    awaymsg = xmlnode_get_tag_data(jp->x, "status");
                    log_debug(ZONE, "[AT] Setting user away");
                    aim_bos_setprofile(s->ass, 
                            aim_getconn_type(s->ass, AIM_CONN_TYPE_BOS),
                            profile, awaymsg, AIM_CAPS_CHAT);
                    /* at_psend(s->mp_to, jp, 0); */
                    x = xmlnode_dup(jp->x);
                    xmlnode_put_attrib(x, "from", jid_full(s->from));
                    xmlnode_put_attrib(x, "to", jid_full(s->cur));
                    log_debug(ZONE, "[AT] Pres Send: %s", xmlnode2str(x));
                    deliver(dpacket_new(x), ti->i);
                    xmlnode_free(jp->x);
					return 1;
                }
				if(jp->to->resource == NULL)
				{
					return at_buddy_add(ti, jp);
				} else {
                    char profile[] = "";
                    /* user is already connected, don't start a new session */
                    /* No show let's reset their profile */
                    log_debug(ZONE, "[AT] Setting user away\n");
                    if(s!=NULL)
                    {
                        aim_bos_setprofile(s->ass, 
                                aim_getconn_type(s->ass, AIM_CONN_TYPE_BOS),
                                profile, NULL, AIM_CAPS_CHAT);
                        x = xmlnode_dup(jp->x);
                        xmlnode_put_attrib(x, "from", jid_full(s->from));
                        xmlnode_put_attrib(x, "to", jid_full(s->cur));
                        log_debug(ZONE, "[AT] Pres Send: %s", xmlnode2str(x));
                        deliver(dpacket_new(x), ti->i);
                        xmlnode_free(jp->x);
                        return 1;
                    }
					return at_auth_user(ti, jp);
				}
                x = xmlnode_dup(jp->x);
				xmlnode_put_attrib(x, "from", jid_full(jp->to));
				xmlnode_put_attrib(x, "to", jid_full(jp->from));
                deliver(dpacket_new(x), ti->i);

                xmlnode_free(jp->x);
                return 1;
        default:
            xmlnode_free(jp->x);
            return 1;
    }

    return 0;
}
